/*
    AFGcockpit
    Arduino FlightGear cockpit

    Hans Kunkell, 2020.
    
    
    
 Flightgear start options:
 --generic=serial,out,4,/dev/ttyUSB0,38400,AFGcockpitout
 --generic=serial,in,100,/dev/ttyUSB0,38400,AFGcockpitin
 
*/

// Use the Streaming library so we can write using << :
#include <Streaming.h>

#include "afgcockpit.h"
#include "analogpots.h"
#include "pushbuttons.h"
#include "encoders.h"
#include "shiftreg.h"
#include "mux.h"
#include "heartbeat.h"
#include "strings.h"
#include "lcd.h"



// This variable becomes true when some input has has been registered/changed.
// We will only then send message to FG, otherwise it will not be able to parse
// everything and become unresponsive.
int sendmsg = false;

// If FG property /Arduino/read-only is 1, Arduino will pause sending messages.
// This is controlled by this variable:
int read_only = 0;


// This are timers to control that we do not send messages to often (if not
// necessary). Set the minimum time between posts with "minwait", it should
// correspond to reading frequency in FG. If minwait=100 [ms] => 10 Hz and
// the start options for the generic protocol should match this (or be higher).
// If the FG frequency is set too low, it will feel sluggish.
// "maxwait" is the time until we send message to FG even if nothing has triggered
// a change.
unsigned long timer1 = 0;
unsigned long timer2 = 0;
int minwait = 10;
int maxwait = 5000;
long mainloop_counter = 0;
long mainloop_frequency = 0;
unsigned long mainloop_timer = 0;

// Counters for messages received and sent:
struct {
  unsigned long fromFG = 0;
  unsigned long toFG = 0;
  unsigned long chunks_fromFG = 0;
  unsigned long chunks_discarded = 0;
} msg;

// Syncing with FG: Every time Arduino sends a message it includes the number (counter)
// of the messages sent and the syncing flag "FG_sync" becomes false. The counter value
// will be stored in FG (/Arduino/msg-to-FG). When FG sends data back to Arduino it also
// includes this counter value which gets stored in msg.toFG variable.
// Back here we can see that if the numbers of messages sent
// are same as received by FG, we are in sync.
int FG_sync = true;


////////////////////////////////////// SYSTEM SETUP

#if NANALOGPOTS > 0
AnalogPot analogpots[NANALOGPOTS];
#endif

#if NPUSHBUTTONS > 0
PushButton pushbuttons[NPUSHBUTTONS];
#endif

#if NENCODERS > 0
Encoder encoder[NENCODERS];
#endif

#if NMUXERCONTROLLERS > 0
MuxAnalog mux[NMUXERCHANNELS];
#endif

#if NSHIFTREGISTERS > 0
ShiftRegister shift_register;
ShiftRegisterSection output[__SHIFTREGISTERLAST__];
#endif

////////////////////////////////////// SYSTEM SETUP ~~~~~~~~~~~~~~~~~~~~~~~



// Write values to FG property tree:


// Read-only values from FG property tree:
// Define maximum length of the strings, this include final \0 byte:
#define LEN_ADFMODE 4
#define LEN_CLOCK 9
#define LEN_ADFRIGHTDISPLAY 5

struct {
  int ap_state = 0;
  int ap_rol = 0;
  int ap_hdg = 0;
  int ap_fpm = 0;
  float altimeter = 0;
  int engine_running = 0;
  float rpm = 0;
  float airspeedindicator = 0;
  float turnindicator = 0;
  float verticalspeedindicator = 0;
  float magneticcompass = 0;
  float headingindicator = 0;
  float headingindicatoroffset = 0;
  float flaps = 0;
  float volts = 0;
  float amps = 0;
  float fuel_left_cap = 0;
  float fuel_left_level = 0;
  float fuel_right_cap = 0;
  float fuel_right_level = 0;
  float oil_pressure = 0;
  float oil_temperature = 0;
  float air_temperature = 0;
  int brake_left = 0;
  int brake_right = 0;
  char clock[LEN_CLOCK] = {'\0'};
  char adf_mode[LEN_ADFMODE] = {'\0'};
  int adf_freq_select = 0;
  int adf_freq_standby = 0;
  char adf_rightdisplay[LEN_ADFRIGHTDISPLAY] = {'\0'};
} fgdata;


void setup(){
  // Initiate serial:
  Serial.begin(SERIALBAUDRATE);

  #ifdef ARDEBUG
  Serial.println("\n\nsetup()");
  #endif

  // Set timer to current time:
  timer1 = millis();
  timer2 = millis();

  #ifdef HEARTBEAT
  setup_heartbeat();
  #endif

  #ifdef SYNCLED
  pinMode(syncled_ledpin, OUTPUT);
  #endif

  // Setup sensors ==========================================================================

  #if NPUSHBUTTONS > 0
  setupPushButtons();
  // add custom settings for pushbuttons here.
  #endif

  #if NANALOGPOTS > 0
  // add custom settings for directly connected analog potentiometers here.
  analogpots[ELEVATOR_TRIM].mode = RELATIVE_VALUE;
  analogpots[ELEVATOR_TRIM].measure_averages = 3;

  setupAnalogPots();
  clearAnalogPots();
  #endif

  #if NMUXERCONTROLLERS > 0
  setupMuxers();
  // add custom settings for muxers here.
  mux[ADFADF].mode = MIMIC_PUSH_BUTTON;
  mux[ADFBFO].mode = MIMIC_PUSH_BUTTON;
  mux[THROTTLE].minout = 0;
  mux[THROTTLE].precision = 3;
//  mux[THROTTLE].delay = 0;  // read value overy loop.
  mux[AILERON].minout = -1;
  mux[AILERON].precision = 3;
  mux[AILERON].measure_averages = 3;
//  mux[AILERON].delay = 0;  // read value overy loop.
  mux[ELEVATOR].minout = -1;
  mux[ELEVATOR].precision = 3;
  mux[ELEVATOR].measure_averages = 3;
//  mux[ELEVATOR].delay = 0;  // read value overy loop.
  mux[ELEVATOR].mode = RELATIVE_VALUE;

  #endif

  // Setup encoders:
  #if NENCODERS > 0
  setupEncoders();
  #endif


  #if NSHIFTREGISTERS > 0
  defineSRdisplay(&output[ADFLEFT], -1, 4);
  defineSRdisplay(&output[ADFRIGHT], -1, 4);
  setupShiftRegisters();
  #endif



  // Show boot/setup finished
  for (int a=0; a<5; a++) {
    digitalWrite(LED_BUILTIN, HIGH);
    delay(250);
    digitalWrite(LED_BUILTIN, LOW);
    delay(100);
  }

  #ifdef ARDEBUG
  Serial.println("setup() finished");
  #endif

}



void loop()
{

  #ifdef HEARTBEAT
  heartbeat();
  #endif

  mainloop_frequency_counter();

  // Handle serial input from FG (if available):
  parseFGdata();

  readSensors();

  #if NENCODERS > 0
  calculateEncoders();
  #endif

  // Send data to FG (if necessary):
  sendtoFG();

  syncledpin();

  #if NSHIFTREGISTERS > 0
  updateShiftRegisters();
  shiftoutShiftRegisters();
  #endif
} // end of main loop



void readSensors() {

  #if NPUSHBUTTONS > 0
  readPushButtons();
  #endif

  #if NANALOGPOTS > 0
  readAnalogPots();
  #endif

  #if NMUXERCONTROLLERS > 0
  readMuxers();
  #endif  
}



void mainloop_frequency_counter() {
  // This is only for the purpose of calculating the frequency of the main loop:
  mainloop_counter++;
  if (mainloop_timer < millis()) {
    mainloop_timer = millis() + 1000;
    mainloop_frequency = mainloop_counter;
    mainloop_counter = 0;
  }
}


void parseFGdata() {

  // Store the the message identifier/tag/name in this variable 'msgid'. Maximum
  // length of the message id is 10-2=8 letters due to NUL character at end and the :.
  char msgid[10];

  // Only : would be enough I guess...
  char msgstopchars[] = { ':', '\n', PROTOCOL_VARSEP, '\0' };

  // See if we have serial data to read:
  while (Serial.available()  >  0) {

    _SerialReadString(msgid, 10, msgstopchars);
     //Serial.print(msgid);

     // Increase counter for received chunks. Each message consists of multiple chunks.
     msg.chunks_fromFG++;

    // Now we know which message identifier is received, we will read the correct type
    // of value and store it in corresponding variable.

    if      (0 == strcmp(msgid, "MSG"))   { HandleMsgNumber(); msg.fromFG++; }
    else if (0 == strcmp(msgid, "RO"))    { read_only = SerialReadInt(); }

    else if (0 == strcmp(msgid, "APA"))  { fgdata.ap_state = SerialReadInt(); }
    else if (0 == strcmp(msgid, "APR"))  { fgdata.ap_rol = SerialReadInt(); }
    else if (0 == strcmp(msgid, "APH"))  { fgdata.ap_hdg = SerialReadInt(); }
    else if (0 == strcmp(msgid, "APF"))  { fgdata.ap_fpm = SerialReadInt(); }

    else if (0 == strcmp(msgid, "ALT"))  { fgdata.altimeter = SerialReadDouble(); }
    else if (0 == strcmp(msgid, "ENG"))  { fgdata.engine_running = SerialReadInt(); }
    else if (0 == strcmp(msgid, "RPM"))  { fgdata.rpm = SerialReadDouble(); }
    else if (0 == strcmp(msgid, "ASI"))  { fgdata.airspeedindicator = SerialReadDouble(); }
    else if (0 == strcmp(msgid, "TUI"))  { fgdata.turnindicator = SerialReadDouble(); }
    else if (0 == strcmp(msgid, "VSI"))  { fgdata.verticalspeedindicator = SerialReadDouble(); }
    else if (0 == strcmp(msgid, "MCO"))  { fgdata.magneticcompass = SerialReadDouble(); }
    else if (0 == strcmp(msgid, "HDI"))  { fgdata.headingindicator = SerialReadDouble(); }
    else if (0 == strcmp(msgid, "HDO"))  { fgdata.headingindicatoroffset = SerialReadDouble(); }
    else if (0 == strcmp(msgid, "FLA"))  { fgdata.flaps = SerialReadDouble(); }
    else if (0 == strcmp(msgid, "VLT"))  { fgdata.volts = SerialReadDouble(); }
    else if (0 == strcmp(msgid, "AMP"))  { fgdata.amps = SerialReadDouble(); }
    else if (0 == strcmp(msgid, "FLC"))  { fgdata.fuel_left_cap = SerialReadDouble(); }
    else if (0 == strcmp(msgid, "FLL"))  { fgdata.fuel_left_level = SerialReadDouble(); }
    else if (0 == strcmp(msgid, "FRC"))  { fgdata.fuel_right_cap = SerialReadDouble(); }
    else if (0 == strcmp(msgid, "FRL"))  { fgdata.fuel_right_level = SerialReadDouble(); }
    else if (0 == strcmp(msgid, "POL"))  { fgdata.oil_pressure = SerialReadDouble(); }
    else if (0 == strcmp(msgid, "TOL"))  { fgdata.oil_temperature = SerialReadDouble(); }
    else if (0 == strcmp(msgid, "TAI"))  { fgdata.air_temperature = SerialReadDouble(); }
    else if (0 == strcmp(msgid, "BRL"))  { fgdata.brake_left = SerialReadInt(); }
    else if (0 == strcmp(msgid, "BRR"))  { fgdata.brake_right = SerialReadInt(); }
    else if (0 == strcmp(msgid, "CLK"))  { SerialReadString(fgdata.clock, LEN_CLOCK); }
    else if (0 == strcmp(msgid, "ADFmode"))  { SerialReadString(fgdata.adf_mode, LEN_ADFMODE); }
    else if (0 == strcmp(msgid, "ADFdisp"))  { SerialReadString(fgdata.adf_rightdisplay, LEN_ADFRIGHTDISPLAY); }
    else if (0 == strcmp(msgid, "ADFfreq1"))  { fgdata.adf_freq_select = SerialReadInt(); }
    else if (0 == strcmp(msgid, "ADFfreq2"))  { fgdata.adf_freq_standby = SerialReadInt(); }


    else if (0 == strcmp(msgid, "DUMP")) { debugDump(); }
    else {
      msg.chunks_discarded++;
      // Removes the character up to next var separator or newline, so
      // that we can continue with next message:
      SerialFlush();
    }

  }
  
}


void sendtoFG() {
  // Update message and send to host computer (FG)

  // If we currently is in "read only" mode, no sending shall occur.
  if (read_only == 1)
    return;

  unsigned long now = millis();
   if ((sendmsg && timer1 < now) || timer2 < now) {
    //Update the two timers:
    timer1 = now+minwait;
    timer2 = now+maxwait;

    // We set sendmsg to false so we will not send any more messages until we
    // have new input (timer2 will still cause messages to send if we wait long enough):
    sendmsg = false;
    // Increase the counter for sent messages and set syncing with FG (FG_sync) to false.
    // Only when FG returns this number (msg.toFG) we will put FG_sync to true again,
    // since we then are in sync.
    msg.toFG++;
    FG_sync = false;

    // Send the messages (this should match the XML protocol).
    Serial.print(msg.toFG);                  Serial.print(",");  //Messages sent
    Serial.print(msg.fromFG);              Serial.print(",");  //Messages received through serial
    Serial.print(msg.chunks_fromFG);       Serial.print(",");  //Messages sent
    Serial.print(msg.chunks_discarded);        Serial.print(",");  //Messages sent
    Serial.print(mainloop_frequency);        Serial.print(",");  //Frequency for main loop (=fps)
    Serial.print("|,");

    #if NPUSHBUTTONS > 0
    printPushButtons();
    #endif
    Serial.print("|,");

    #if NANALOGPOTS > 0
    printAnalogPots();
    clearAnalogPots();
    #endif
    Serial.print("|,");

    #if NMUXERCONTROLLERS > 0
    printMuxers();
    clearMuxers();
    #endif
    Serial.print("|,");

    #if NENCODERS > 0
    printEncoders(",\0");
    clearEncoders();
    #endif
    Serial.print("|,");

    // uptime:
    Serial.print(millis());

    Serial.print("\n");

  }


  

}

void updateShiftRegisters() {
  if (shift_register.timer > millis()+shift_register.update_delay)
    return;

  // Define here which variables should match which shift registers.

  writeShiftRegisterInt(&output[ADFLEFT], fgdata.adf_freq_select);
  writeShiftRegisterStr(&output[ADFRIGHT], fgdata.adf_rightdisplay);



}


//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//Additional and general functions
// Toggles a variable between 0 and 1:
int toggle(int val) {
  if (val == 1)
    val = 0;
  else
    val = 1;
  return(val);
}

void syncledpin() {
  #ifdef SYNCLED
    if (FG_sync)  digitalWrite(syncled_ledpin, HIGH);
    else          digitalWrite(syncled_ledpin, LOW);
  #endif  
}

// These are used for ascii serial data. It will read the message number and if it is
// the same as the number sent from Arduino, we will set our sync flag to true!
void HandleMsgNumber() {
  long id = 0;
  //Read the number, which is a long integer:
  id = SerialReadLong();
  if (id == msg.toFG) {
    FG_sync = true;
  }
}




double mapf(float x, float in_min, float in_max, float out_min, float out_max)
{
  //This function is same as the builtin map function but this is with floats.
  return (x - in_min) * (out_max - out_min) / (in_max - in_min) + out_min;
}

int OutsideRange(float val, float old, float range) {
  // This function returns true if a value is outside the range of a previous value +/-
  // a neighbouring area. If we are inside this area, which corresponds to no change in 
  // read values, we return false.
  if (val < 0.0) {
    if (val < old-range || val > old+range)
      return(true);
  }
  else {
    if (val > old+range || val < old-range) 
      return(true);
  }

  return(false);
}


// This functions takes a value usually read by analogWrite and converts it to discrete values
// defined by a number of steps.
int analog2discrete(int value, int minv, int maxv, int steps) {
  // Calculate the range for each value:
  int range = (maxv-minv) / steps;
  int counter = 1;
  // Loop as long as our value is larger than a specific range:
  while (value > minv+counter*range)
    counter++;
  
  // To eliminate rounding errors that may cause the counter to be larger
  // than the number of steps, we limit it here:
  if (counter > steps)
    counter = steps;
  return(counter);
}


void debugDump() {

  Serial.print("ADF mode="); Serial.println(fgdata.adf_mode);
  Serial.print("ADF display="); Serial.println(fgdata.adf_rightdisplay);
  Serial.print("Clock="); Serial.println(fgdata.clock);
  printShiftRegistersToSerial();
}
