// strings.cpp

// Functions to handle strings and some Serial data.

#include "afgcockpit.h"
#include "strings.h"

// Return true if strings are equal. Also
// works on strings that are not null-terminated.
bool strequal(char *s1, char *s2, int length) {
  for (int i=0; i<length; i++) {
    if (s1[i] != s2[i])
      return false;
  }
  return true;
}


// Removes the remaining string (until newline or variable separator) from serial data:
void SerialFlush() {
  char str = '\0';
  // If this function is called twice this prevents it from hanging until we have
  // serial data available:
  if (Serial.available() == 0)
    return;

  while(1<2) {
    if (Serial.available() > 0) {
      str = Serial.read();
      if (str == '\n' || str == PROTOCOL_VARSEP)
        return;
    }
    else
      delayMicroseconds(SERIALREADWAIT);
  }
}

void SerialReadString(char *str, int maxlen) {
  char stopchars[] = { '\n', PROTOCOL_VARSEP, '\0' };
  _SerialReadString(str, maxlen, stopchars);
}

void _SerialReadString(char *str, int maxlen, char *stopchars) {
  // Read data from Serial connection. Stop reading when any of the characters found in
  // stopchars is matched. That is in general a newline, NUL, PROTOCOL_VARSEP.
  // Put a NUL at the end. Do not write more than 'maxlen' (that includes final
  // NUL character).
  int c=0;

  // Add +1 to include check for NUL:
  int stops = strlen(stopchars)+1;
  bool enough = false;
  while( !enough ) {
    // See if we have serial data available:
    if (Serial.available() > 0) {
      // Yes, we have!
      // Store it at next position in string:
      str[c] = Serial.read();
      // If it is newline or we find a variable separator, then terminate the string
      // and leave the otherwise infinite loop:
      for (int i=0; i<stops; i++) {
        if (str[c] == stopchars[i]) {
          str[c] = '\0';
          enough = true;
        }
      }

      if (!enough) {
        // Ok, we were not at the end of the string, go on with next:
        c++;

        // Remove any chars that cannot fit into *str. maxlen does not include NUL.
        if (c == maxlen-1) {
          str[c] = '\0';
          SerialFlush();
          break;
        }
      }
    } /* Serial.available() */
    // In case there are not serial data available, make a short pause before retrying:
    else
      delayMicroseconds(SERIALREADWAIT);
  } /* while */
}

int SerialReadInt() {
  // The string we read from Serial will be stored here:
  char str[16];
  SerialReadString(str, 16);
  // Convert the string to int and return:
  return(atoi(str));
}

long SerialReadLong() {
  char str[16];
  SerialReadString(str, 16);
  return(atol(str));
}

double SerialReadDouble() {
  char str[16];
  SerialReadString(str, 16);
  return(atof(str));
}


void PrintDouble( double val, byte precision){
  // prints val with number of decimal places determine by precision
  // precision is a number from 0 to 6 indicating the desired decimial places
  // example: lcdPrintDouble( 3.1415, 2); // prints 3.14 (two decimal places)
  // From http://www.arduino.cc/cgi-bin/yabb2/YaBB.pl?num=1207226548

  if(val < 0.0){
    Serial.print('-');
    val = -val;
  }

  Serial.print (int(val));  //prints the int part
  if( precision > 0) {
    Serial.print("."); // print the decimal point
    unsigned long frac;
    unsigned long mult = 1;
    byte padding = precision -1;
    while(precision--)
  mult *=10;

    if(val >= 0)
 frac = (val - int(val)) * mult;
    else
 frac = (int(val)- val ) * mult;
    unsigned long frac1 = frac;
    while( frac1 /= 10 )
 padding--;
    while(  padding--)
 Serial.print("0");
    Serial.print(frac,DEC) ;
  }
}

int ReadAnalogDiscreteBool(int pin, int var) {
  int tmpi = analog2discrete(analogRead(pin), 0, 1023, 2) - 1;
  if (tmpi != var)
    sendmsg = true;
  return(tmpi);
}

float ReadAnalogDiscreteFloat(int pin, float var, int steps, float div, float red, float range) {
  int tmpi = analogRead(pin);
  float tmpf = (float) analog2discrete(tmpi, 0, 1023, steps) / div - red;
  if (OutsideRange(tmpf, var, range))
    sendmsg = true;
  return(tmpf);
}



int num2str(char *str, double val, byte precision){
  // This function converts a number (double) to a string. To only consider integers,
  // precision should be set to 0. The length of the string is returned (excl. '\0').
  // There are occasionally problems with the fractions. The cause: Rounding errors.
  // In a future version could the last digit be studied, if it's a 9, increase all preceeding
  // 9's, like 111.699 becomes 111.70 instead of 111.69.

  int pos = 0;
  int i;
  long test;
  long num;
  int isinside=false;
  const int ntest = 8;

  
  // First deal with negative numbers:
  if(val < 0.0){
    str[pos] = '-';
    pos++;
    val = -val;
  }
  else if (long(val) == 0) {
    str[pos] = '0';
    pos++;
  }

  // Print the integer part:
  long tester[ntest] = {1,10,100,1000,10000,100000,1000000,10000000};
  for (i=ntest-1; i>=0; i--) {
    test = tester[i];
    if (test <= val) {
      isinside=true;
      num = long(val/test);
      str[pos] = num+'0';
      pos++;
      val = val-num*test;
    }
    else if (isinside) {
      str[pos] = '0';
      pos++;
    }
  }
  
  
  // Print the fraction:
  if( precision > 0) {
    str[pos] = '.';
    pos++;
    while (precision--) {
      val *= 10;
      num = long(val);
      val = val-num;        
      str[pos] = num+'0';
      pos++;
    }
  }

  str[pos] = '\0';

//  Serial.print(str);
//  Serial.print("\n");
  return(pos);
}

//void PrintSR(int start, int len, double number, byte precision) {
//  // Prints a number to the global shift register string SR at a certain position (starts
//  // at "start") and length ("len"). Now it's right aligned.
//  // If the number has more letters than available space (length), the last character will 
//  // be displaying '_' to show it was exceeded.
//  
//  char temp[20];
//  int srcpos=0;
//  int trgpos=0;
//  int nchars;
//  
//  // Put number into string:
//  nchars = num2str(temp, number, precision);
//  // If we have fraction, the . takes up one byte but it should be placed within adjacent unit number.
//  if (precision > 0)
//    nchars = nchars-1;
//
//  // Pad with spaces before:  
//  while (len-trgpos > nchars) {
//    SR[start+trgpos] = char2LED7(' ');
//    trgpos++;
//  }
//  
//  // Copy converted characters to SR string:
//  while (trgpos<len) {
//    SR[start+trgpos] = char2LED7(temp[srcpos]);
//    if (temp[srcpos+1] == '.') {
//      bitClear(SR[start+trgpos], 7);
//      srcpos++;
//    }
//    
//    trgpos++;
//    srcpos++;
//  }
//
//
//  // If overflow of letters, inform:
//  if (nchars > len)
//    SR[start+len-1] = char2LED7('_');
//}
